function [r d Q] = BuildHWtree(R,P,jMax,N,x,dx,dt,pu,pm,pd);

% Construct the Hull-White interest rate tree using the simplified
% approach.  Inputs are:
% R    = zero coupon rates
% P    = Discount factors.
% jMax = Number of up/down steps.
% N    = Number of time steps
% x    = Auxiliary state variable.
% dx   = State varible increment.
% dt   = Time increment.

% Initialize the required matrices.
r  = zeros(2*jMax+1,N+1);
d  = zeros(2*jMax+1,N+1);
Q  = zeros(2*jMax+1,N+1);

% J vector used to obtain probabilities pu, pm, pd
J = [jMax:-1:-jMax];

% First Arrow-Debreu price
Q(jMax+1,1) = 1;

% Build the interest rate tree.  Calculate separately:
%   - the first node (Column 1)
%   - the "tip" of the tree (Columns 2 through jMax +1)
%   - the "box" of the tree (Columns Jmax+2 through N)
for j=1:N+1
	if j==1
		% First node: column 1
		Q(jMax+1,1) = 1;
		d(jMax+1,1) = exp(-R(1)*dt);
		a(1) = -log(P(1))/dt;
		S(1) = Q(jMax+1,1)*exp(-x(jMax+1)*dx*dt);
		r(jMax+1,1) = x(jMax+1) + a(1);
	elseif j<=jMax+1;
		% Tip of the tree: columns 2 through jMax+1
		for i=jMax-(j-2):jMax+j
			if i==jMax-(j-2)        % First row
				Q(i,j) = Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
			elseif i==jMax-(j-2)+1  % Second row
				Q(i,j) = Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
			elseif i==jMax+j        %Last row
				Q(i,j) = Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1);
			elseif i==jMax+j-1      % Second to last row
				Q(i,j) = Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1);
			else                    % Middle rows
				Q(i,j) = Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
			end
			S(j) = 0;
			for k=jMax-(j-2):jMax+j
				S(j) = S(j) + Q(k,j)*exp(-J(k)*dx*dt);
			end
			a(j) = (log(S(j)) - log(P(j)))/dt;
			for k=jMax-(j-2):jMax+j
				r(k,j) = x(k) + a(j);
				d(k,j) = exp(-r(k,j)*dt);
			end
		end
	else
		% Box of the tree: columns jMax+2 through N
		for i=1:2*jMax+1
			if i==1             % First row
				Q(i,j) = Q(i,j-1)*pu(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);;
			elseif i==2         % Second row
				Q(i,j) = Q(i-1,j-1)*pm(i-1,j-1)*d(i-1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
			elseif i==3         % Third row
				if jMax==2      % Special case of third row when jMax=2
					Q(i,j) = Q(i-2,j-1)*pd(i-2,j-1)*d(i-2,j-1) + Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1) + Q(i+2,j-1)*pu(i+2,j-1)*d(i+2,j-1);
				else
					Q(i,j) = Q(i-2,j-1)*pd(i-2,j-1)*d(i-2,j-1) + Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
				end
			elseif i==2*jMax+1  % Last row
				Q(i,j) = Q(i,j-1)*pd(i,j-1)*d(i,j-1) + Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1);
			elseif i==2*jMax    % Second to last row
				Q(i,j) = Q(i+1,j-1)*pm(i+1,j-1)*d(i+1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1);
			else                % Middle rows
				Q(i,j) = Q(i-1,j-1)*pd(i-1,j-1)*d(i-1,j-1) + Q(i,j-1)*pm(i,j-1)*d(i,j-1) + Q(i+1,j-1)*pu(i+1,j-1)*d(i+1,j-1);
			end
			S(j) = 0;
			for k=1:2*jMax+1
				S(j) = S(j) + Q(k,j)*exp(-J(k)*dx*dt);
			end
			a(j) = (log(S(j)) - log(P(j)))/dt;
			for k=1:2*jMax+1
				r(k,j) = x(k) + a(j);
				d(k,j) = exp(-r(k,j)*dt);
			end
		end
	end
end
