% Generates the Hull White trinomial tree, using the simplified formulation
% described in Section 9.5 of Clewlow and Strickland 
% "Implementing Derivatives Models"

clc; clear;
% Tenors (1w 2w 3w 1m 2m 3m 4m 5m 6m 9m 1y 1.5y 2y 2.5y 3y 4y)
Tenors = [1/52 2/52 3/52 1/12 2/12 3/12 4/12 5/12 6/12 9/12 1 1.5 2 2.5 3 4];

% Corresponding zero coupon rates
Rates = [3 3.1 3.15 3.19 3.3 3.5 3.71 3.89 4.05 4.15 4.35 4.66 4.86 5.06 5.15 5.76]./100;

% Maturity in years (T) number of time steps (N) and time increment (dt)
T = 3;
N = 10;
dt = T/N;
Time = dt:dt:T+dt;

% Start and end time of each time interval
TimeIntervals = [0:dt:T; Time]

% Impute R(1) when dt < Tenor(1)
R = interp1(Tenors, Rates, Time);
R(find(isnan(R))) = Rates(1);

% Discount factor.
P = exp(-R.*Time);

% Strike price for puts on the discount bond
K = 0.95;

% Hull White tree parameters
alpha = 0.1;
sigma = 0.005;
dr = sigma*sqrt(3*dt);
dx = dr;

% Threshold (M), jMax, and state vector (x)
M = -alpha*dt;
jMax = ceil(-0.1835/M);
x = dr.*[jMax:-1:-jMax];

% Construct the trees for the probabilities pu, pm, pd.
[pu pm pd] = BuildHWprobs(jMax,N,M);

% Construct the tree for interest rates (r), discount factors (d),
% and Arrow-Debreu securities.
[r d Q] = BuildHWtree(R,P,jMax,N,x,dx,dt,pu,pm,pd);

HWTree = r.*100

% Initialize the discount bond B.
B = zeros(2*jMax+1,N+1);

% Last column of discount bond is 1.
B(:,end) = 1;

% Work backwards through tree to get remaining discount bond prices
for j=N:-1:1
	if j>jMax
		for i=1:2*jMax+1
			if i==1             % first row.
				B(i,j) = d(i,j)*(B(i,j+1)*pu(i,j) + B(i+1,j+1)*pm(i,j) + B(i+2,j+1)*pd(i,j));
			elseif i==2*jMax+1  % last row.
				B(i,j) = d(i,j)*(B(i,j+1)*pd(i,j) + B(i-1,j+1)*pm(i,j) + B(i-2,j+1)*pu(i,j));
			else
				B(i,j) = d(i,j)*(B(i-1,j+1)*pu(i,j) + B(i,j+1)*pm(i,j) + B(i+1,j+1)*pd(i,j));
			end
		end
	else
		for i=jMax-(j-2):jMax+j
			B(i,j) = d(i,j)*(B(i-1,j+1)*pu(i,j) + B(i,j+1)*pm(i,j) + B(i+1,j+1)*pd(i,j));
		end
	end
end

% Print out the Discount bond (middle of Figure 9.17)
B;

% Price the American and European puts.  
% Maturity is three years.
Mat = 3;

% European put is the payoff times the Arrow Debreu prices
Payoff = max(K - B(:,Mat+1), 0);
EuroPut = Q(:,Mat+1)' * Payoff

% Initialize the American puts
AP = zeros(2*jMax+1,Mat+1);

% Intrinsic value at maturity.
AP(:,Mat+1) = max(K-B(:,Mat+1),0);
EP = AP;

% Work backwards through the tree
for j=Mat:-1:1
	if j>jMax   % Box of the tree
		for i=1:2*jMax+1
			if i==1             % first row.
				EP(i,j) = d(i,j)*(EP(i,j+1)*pu(i,j) + EP(i+1,j+1)*pm(i,j) + EP(i+2,j+1)*pd(i,j));
			elseif i==2*jMax+1  % last row.
				EP(i,j) = d(i,j)*(EP(i,j+1)*pd(i,j) + EP(i-1,j+1)*pm(i,j) + EP(i-2,j+1)*pu(i,j));
			else                % middle rows
				EP(i,j) = d(i,j)*(EP(i-1,j+1)*pu(i,j) + EP(i,j+1)*pm(i,j) + EP(i+1,j+1)*pd(i,j));
			end
			AP(i,j) = max(EP(i,j), K - B(i,j));
		end
	else       % Tip of the tree
		for i=jMax-(j-2):jMax+j
			EP(i,j) = d(i,j)*(EP(i-1,j+1)*pu(i,j) + EP(i,j+1)*pm(i,j) + EP(i+1,j+1)*pd(i,j));
			AP(i,j) = max(EP(i,j), K - B(i,j));
		end
	end
end

% Print out the American put tree.
AP;

% Print out the American put price
AmerPut = AP(jMax+1,1)
